from typing import Union
from RPA.Windows.keywords import (
    ActionNotPossible,
    keyword,
    LibraryContext,
    WindowControlError,
)
from RPA.Windows import utils

if utils.is_windows():
    import uiautomation as auto
    from uiautomation.uiautomation import Control


class ControlKeywords(LibraryContext):
    """Keywords for handling Control objects"""

    @keyword
    def print_control_tree(
        self, target=None, max_depth=2, encoding="utf-8", level="info"
    ):
        """[summary]

        :param target: [description], defaults to None
        :param max_depth: [description], defaults to 2
        :param encoding: [description], defaults to "utf-8"
        :param level: [description], defaults to "info"
        """
        index = 1

        def GetFirstChild(control):
            return control.GetFirstChildControl()

        def GetNextSibling(control):
            return control.GetNextSiblingControl()

        # auuto.GetRootControl()
        target = target or self.ctx.window
        for control, depth in auto.WalkTree(
            target,
            getFirstChild=GetFirstChild,
            getNextSibling=GetNextSibling,
            includeTop=True,
            maxDepth=max_depth,
        ):
            control_as_text = (
                str(control).encode(encoding) if encoding else str(control)
            )
            if level == "info":
                self.ctx.logger.info(f"{' ' * depth * 4}{control_as_text}")
            else:
                self.ctx.logger.warning(f"{' ' * depth * 4}{control_as_text}")
            # control.CaptureToImage(f"{control.ControlType}_{index}.png")
            index += 1

    @keyword
    def get_control_attribute(self, locator: Union[str, Control], attribute: str):
        """Get attribute of Control object matching locator.

        :param locator: string locator or Control object
        :param attribute: name of the attribute to get
        """
        control = locator
        if isinstance(locator, str):
            try:
                control = self.ctx.get_control(locator)
            except Exception as err:
                raise WindowControlError(str(err)) from err
        if hasattr(control, attribute):
            attr = getattr(control, attribute, None)
            return attr() if callable(attr) else str(attr)
        else:
            raise ActionNotPossible(
                "Control '%s' does not have '%s' attribute" % (locator, attribute)
            )



from typing import Union, Optional

from RPA.Windows.keywords import (
    ActionNotPossible,
    keyword,
    LibraryContext,
    WindowControlError,
)
from RPA.Windows import utils

if utils.is_windows():
    import uiautomation as auto
    from uiautomation.uiautomation import Control


class ActionKeywords(LibraryContext):
    """Keywords for performing desktop actions"""

    @keyword(tags=["action"])
    def click(
        self,
        locator: Union[str, Control],
        set_focus: bool = False,
    ):
        """Mouse click on control matching given locator.

        :param locator: string locator or Control object
        :param set_focus: set True to focus on Control object
        """
        control = locator
        if isinstance(locator, str):
            try:
                control = self.ctx.get_control(locator)
            except Exception as err:
                raise WindowControlError(str(err)) from err
        if set_focus:
            control.SetFocus()
            if hasattr(control, "SetActive"):
                control.SetActive()
        control.MoveCursorToMyCenter(simulateMove=self.ctx.simulate_move)
        if control.robocorp_click_offset:
            rect = control.BoundingRectangle
            offset_x, offset_y = [
                int(v) for v in control.robocorp_click_offset.split(",")
            ]
            auto.Click(
                rect.xcenter() + offset_x,
                rect.ycenter() + offset_y,
                waitTime=self.ctx.timeout,
            )
        else:
            control.Click(
                waitTime=self.ctx.timeout, simulateMove=self.ctx.simulate_move
            )
        return control

    @keyword(tags=["action"])
    def select(self, locator: Union[str, Control], value: str):
        """Select value on Control object if action is supported.

        Will print warning if it is not possible to select on object.

        :param locator: string locator or Control object
        :param value: string value to select on Control object
        """
        control = locator
        if isinstance(locator, str):
            try:
                control = self.ctx.get_control(locator)
            except Exception as err:
                raise WindowControlError(str(err)) from err
        if hasattr(control, "Select"):
            control.Select(value)
        else:
            raise ActionNotPossible(
                "Control '%s' does not have 'Select' attribute" % locator
            )

    @keyword(tags=["action"])
    def input_text(self, text: str, locator: Union[str, Control]):
        """Input text into Control object defined by the locator.

        :param text: the text to input
        :param locator: string locator or Control object
        """
        control = locator
        if isinstance(locator, str):
            control = self.ctx.get_control(locator)
        self.send_keys(text, control)

    @keyword(tags=["action"])
    def send_keys(self, keys: str, locator: Optional[Union[str, Control]] = None):
        """Send keys to desktop, current window or to Control object
        defined by given locator.

        :param keys: the keys to send
        :param locator: string locator or Control object (default None)
        """
        if not locator:
            control = self.ctx.window or auto
        elif isinstance(locator, str):
            control = self.ctx.get_control(locator)
        else:
            control = locator
        control.SendKeys(keys, waitTime=self.ctx.timeout)

    @keyword
    def get_text(self, locator: Union[str, Control]) -> str:
        """Get text from Control object defined by the locator.

        :param locator: string locator or Control object
        """
        control = locator
        if isinstance(locator, str):
            control = self.ctx.get_control(locator)
        if hasattr(control, "GetWindowText"):
            return control.GetWindowText()
        raise ActionNotPossible(
            "Control '%s' does not have 'GetWindowText' attribute" % locator
        )

    @keyword
    def get_item_value(self, locator: Union[str, Control]) -> str:
        """Get Control object item value

        :param locator: string locator or Control object
        """
        control = locator
        if isinstance(locator, str):
            control = self.ctx.get_control(locator)
        if hasattr(control, "GetValuePattern"):
            value_pattern = control.GetValuePattern()
            return value_pattern.Value
        raise ActionNotPossible(
            "Control '%s' does not have 'GetValuePattern' attribute" % locator,
        )

    @keyword(tags=["action"])
    def set_item_value(self, locator: Union[str, Control], value: str) -> None:
        """Set Control object item value

        :param locator: string locator or Control object
        :param value: string value to set
        """
        control = locator
        if isinstance(locator, str):
            control = self.ctx.get_control(locator)
        if hasattr(control, "GetValuePattern"):
            value_pattern = control.GetValuePattern()
            value_pattern.SetValue(value)
        else:
            raise ActionNotPossible(
                "Control '%s' does not have 'GetValuePattern' attribute" % locator,
            )

    @keyword
    def set_timeout(self, timeout: float) -> None:
        """Set library timeout for action keywords

        :param timeout: float value (in seconds), e.g. `0.1`
        """
        self.ctx.timeout = timeout



class ControlNotFound(ValueError):
    """No matching controls were found."""


class MultipleControlsFound(ValueError):
    """Multiple matching controls were found, but only one was expected."""


class TimeoutException(ValueError):
    """Timeout reached while waiting for condition."""


class WindowControlError(ValueError):
    """Matching window was not found"""


class ActionNotPossible(ValueError):
    """Action is not possible for the given Control"""


class LibraryContext:
    """Shared context for all keyword libraries."""

    def __init__(self, ctx):
        self.ctx = ctx

    @property
    def logger(self):
        return self.ctx.logger

    @property
    def window(self):
        return self.ctx.window



import logging
import platform
import warnings

# pylint: disable=wrong-import-position
from robotlibcore import DynamicCore


from RPA.Windows.keywords import (
    ActionKeywords,
    ControlKeywords,
    LocatorKeywords,
    RecorderKeywords,
    WindowKeywords,
)


class Windows(DynamicCore):
    """The `Windows` is a libraru that can be used for Windows desktop automation.

    This library is at this moment in "BETA" stage as an alternative
    library for `RPA.Desktop.Windows`. Main difference being that this
    library is using `uiautomation`_ dependency instead of `pywinauto`.

    .. _uiautomation: https://github.com/yinkaisheng/Python-UIAutomation-for-Windows

    **About terminology**

    The most used term in uiautomation package is "Control" which in the library context
    maps to objects of the application like for example Window, Button or ListItem.

    **Locators**

    Locators are based on different strategies that can used identify Control object.

    Available strategies that can be used for the locators:

    =============== =======================
    Key             Maps to search property
    =============== =======================
    name            Name
    class           ClassName
    type            ControlType
    control         ControlType
    id              AutomationId
    automationid    AutomationId
    regex           RegexName
    subname         SubName
    index           foundIndex
    offset          offset coordinates (x (int), y (int)) from control center
    executable      target window by its executable name
    desktop         target desktop
    process         target window by its executable's process id
    =============== =======================

    Getting control over application object (Control) is done usually by first executing
    keyword ``Control Window`` which by default will search from desktop ControlType
    WindowControl objects with the given locator.

    Keyword ``Control Window`` can also be used to target application child WindowControl objects.

    Keyword examples:

    .. code-block:: robotframework

        Control Window    name:Calculator
        Control Window    Calculator  # will execute search by 'name:Calculator'
        Control Window    executable:Spotify.exe

    some example locators:

    .. code-block:: bash

        id:clearButton
        type:Group and name:'Number pad' > type:Button and index:4
        type:Group and name:'Number pad' > control:Button index:5 offset:370,0
        id:Units1 > name:${unit}
        class:Button

    **How to inspect**

    Most common and recommended by Microsoft, inspector tool for Windows, is `Accessibility Insights`_ that
    can be installed separately. Other options are tools `Inspect Object`_  and `UI Automation Verify`_, which
    can be accessed by installing Windows SDK.

    .. _Accessibility Insights: https://github.com/yinkaisheng/Python-UIAutomation-for-Windows
    .. _Inspect Object: https://docs.microsoft.com/en-us/windows/win32/winauto/inspect-objects
    .. _UI Automation Verify: https://docs.microsoft.com/en-us/windows/win32/winauto/ui-automation-verify

    **Recording**

    Library provides some rudimentary keywords for inspecting and recording locators for
    Control objects.

    Keyword ``Inspect Element`` will return list of steps that can be used to interact with Control object.

    Recording can be activated by ``Start Recording`` which then activate mode where upon mouse click
    Control under mouse pointer will be inspected. Recording can be stopped at any time by keyboard `ESC`.
    After recording has been stopped, the recorded steps can be retrieved by keyword ``Get Recording``.

    Example task for the recording

    .. code-block:: robotframework

        *** Tasks ***
        Record Task
            Log To Console    \nRecord steps for the Task
            Start Recording
            ${recording}=    Get Recording    sleeps=${FALSE}
            Log To Console    \n${recording}

    **Examples**

    Both Robot Framework and Python examples follow.

    The library must be imported first.

    .. code-block:: robotframework

        *** Settings ***
        Library    RPA.Windows

    Windows Calculator steps

    .. code-block:: robotframework

        *** Keywords ***
        Do some calculations
            Control Window    name:Calculator
            Click    id:clearButton
            Send Keys   96+4=
            ${result}=    Get Control Property    id:CalculatorResults    name
            Log To Console    \n${result}

    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(self):
        # , timeout: float = None, simulate_move: bool = False
        self.logger = logging.getLogger(__name__)
        self.timeout = 0.5
        self.simulate_move = False
        self.window = None

        # Register keyword libraries to LibCore
        libraries = [
            ActionKeywords(self),
            ControlKeywords(self),
            LocatorKeywords(self),
            RecorderKeywords(self),
            WindowKeywords(self),
        ]
        super().__init__(libraries)



from dataclasses import dataclass, field
import re
from typing import List
from RPA.Windows.keywords import keyword, LibraryContext, WindowControlError
from RPA.Windows import utils

if utils.is_windows():
    import uiautomation as auto
    from uiautomation.uiautomation import Control


WINDOWS_LOCATOR_STRATEGIES = {
    "automationid": "AutomationId",
    "id": "AutomationId",
    "class": "ClassName",
    "control": "ControlType",
    "depth": "depth",
    "name": "Name",
    "regex": "RegexName",
    "subname": "SubName",
    "type": "ControlType",
    "index": "foundIndex",
    "offset": "offset",
    "desktop": "desktop",
    "process": "process",
    "executable": "executable",
}


@dataclass
class MatchObject:
    """Represents all locator parts as object properties"""

    match_type: str = field(default="all")
    match_index: int = None
    locators: List = field(default_factory=list)
    _classes: List = field(default_factory=list)
    regex: str = None
    regex_field: str = None
    max_level: int = 0

    def parse_locator(self, locator: str):
        locator_tree = [loc.strip() for loc in locator.split(">")]
        # self.logger.warning(locator_tree)
        regex = rf"({':|'.join(WINDOWS_LOCATOR_STRATEGIES.keys())}:|or|and)('{{1}}(.+)'{{1}})|(\S+)?"  # noqa: E501
        match_object = MatchObject()
        for level, branch in enumerate(locator_tree):
            parts = re.finditer(regex, branch, re.IGNORECASE)

            default_value = []
            strategy = None

            for part in parts:
                self.handle_locator_part(
                    level, part, match_object, default_value, strategy
                )
            if not strategy and len(default_value) > 0:
                match_object.add_locator("Name", " ".join(default_value), level)
        if len(match_object.locators) == 0:
            match_object.add_locator("Name", locator)
        return match_object

    def handle_locator_part(self, level, part, match_object, default_value, strategy):
        part_text = part.group(0)
        if len(part_text.strip()) == 0:
            return
        if part_text == "or":
            match_object.match_type = "any"
        elif part_text == "and":
            pass
        else:
            try:
                strategy, value = part_text.split(":", 1)
            except ValueError:
                strategy = None
                default_value.append(part_text)
            # self.logger.info("STRATEGY: %s VALUE: %s" % (strategy, value))
            if strategy:
                if len(default_value) > 0:
                    match_object.add_locator("Name", " ".join(default_value))
                    default_value = []
                windows_locator_strategy = WINDOWS_LOCATOR_STRATEGIES[strategy]
                match_object.add_locator(windows_locator_strategy, value, level)

    def add_locator(self, strategy, value, level=0) -> None:
        if not value:
            return
        self.max_level = max(self.max_level, level)
        value = value.replace("'", "").strip()
        if strategy == "regex":
            self.regex = value
        elif strategy == "regex_field":
            self.regex_field = value
        elif strategy == "foundIndex":
            self.match_index = int(value.strip())
            value = int(value.strip())
            self.locators.append([strategy, value, level])
        elif strategy == "ControlType":
            value = value if value.endswith("Control") else f"{value}Control"
            self.locators.append([strategy, value, level])
        else:
            self.locators.append([strategy, value, level])
        if (
            strategy in ["class", "class_name", "friendly", "friendly_class_name"]
            and value.lower() not in self._classes
        ):
            self._classes.append(value.lower())

    @property
    def classes(self) -> List:
        uniques = []
        for c in self._classes:
            if c not in uniques:
                uniques.append(c)
        return uniques


class LocatorKeywords(LibraryContext):
    """Keywords for handling Windows locators"""

    def _get_control_with_locator_part(
        self, locator, search_depth, root_control=None
    ) -> Control:
        control = None

        match_object = MatchObject()
        mo = match_object.parse_locator(locator)
        self.ctx.logger.info("locator '%s' to match object: %s" % (locator, mo))
        search_params = {}
        for loc in mo.locators:
            search_params[loc[0]] = loc[1]
        offset = search_params.pop("offset", None)
        if "executable" in search_params.keys():
            root_control = auto.GetRootControl()
            search_params.pop("ControlType")
            executable = search_params.pop("executable")
            window_list = self.ctx.list_windows()
            matches = [w for w in window_list if w["name"] == executable]
            if not matches:
                raise WindowControlError(
                    "Could not locate window with executable '%s'" % executable
                )
            elif len(matches) > 1:
                raise WindowControlError(
                    "Found more than one window with executable '%s'" % executable
                )
            self.logger.info(
                "Found process with window title: '%s'" % matches[0]["title"]
            )
            search_params["Name"] = matches[0]["title"]
            control = Control(**search_params, searchDepth=search_depth)
            new_control = Control.CreateControlFromControl(control)
            new_control.robocorp_click_offset = offset
            return new_control
        if "desktop" in search_params.keys():
            root_control = auto
            search_params.pop("desktop")
        if "ControlType" in search_params.keys():
            control_type = search_params.pop("ControlType")
            control = getattr(root_control, control_type)
            new_control = control(**search_params, searchDepth=search_depth)
            new_control.robocorp_click_offset = offset
            return new_control

        control = getattr(root_control, "Control")

        new_control = control(**search_params, searchDepth=search_depth)
        new_control.robocorp_click_offset = offset
        return new_control

    @keyword
    def get_control(
        self, locator: str, search_depth: int = 8, root_control: Control = None
    ) -> Control:
        """Get Control object defined by the locator.

        :param locator: locator as a string
        :param search_depth: how deep the Control search will traverse (default 8)
        :param root_control: can be used to restrict Control search into
         a specific Control object
        """
        self.logger.info("Locator '%s' into control", locator)
        locators = locator.split(" > ")
        root_control = root_control or self.ctx.window or auto
        control = None
        for loc in locators:
            self.logger.info("Root control: '%s'" % root_control)
            control = self._get_control_with_locator_part(
                loc, search_depth, root_control
            )
            root_control = control
        self.logger.info("Returning control: '%s'", control)
        return control



from typing import Dict
import platform
import psutil


def get_process_list() -> Dict:
    """Get process list.

    Returns dictionary mapping process id to process name
    """
    process_list = {}
    for proc in psutil.process_iter():
        process_list[proc.pid] = proc.name()
    return process_list


def is_integer(value):
    try:
        float(value)
    except ValueError:
        return False
    else:
        return float(value).is_integer()


def is_windows():
    return platform.system() == "Windows"



import time
from typing import List, Dict

from RPA.Windows.keywords import (
    keyword,
    LibraryContext,
    WindowControlError,
)
from RPA.Windows import utils

if utils.is_windows():
    import uiautomation as auto


class WindowKeywords(LibraryContext):
    """Keywords for handling Window controls"""

    @keyword(tags=["window"])
    def control_window(self, locator: str = None) -> int:
        """Control window defined by the locator string.

        Returns process id of the window.

        :param locator: string locator
        """
        self.ctx.window = self.ctx.get_control(
            locator + " and type:WindowControl", root_control=auto
        )
        if not self.ctx.window.Exists():
            self.ctx.window = self.ctx.get_control(
                locator + " and type:PaneControl", root_control=auto
            )
        if not self.ctx.window.Exists():
            raise WindowControlError(
                "Could not locate window with locator '%s'" % locator
            )
        if hasattr(self.ctx.window, "Restore"):
            self.ctx.window.Restore()
        self.ctx.window.SetFocus()
        self.ctx.window.MoveCursorToMyCenter(simulateMove=self.ctx.simulate_move)
        time.sleep(1.0)
        return self.ctx.window.ProcessId

    @keyword(tags=["window"])
    def minimize_window(self, locator: str = None) -> None:
        """Minimize window defined by the locator string or the
        current active window.

        :param locator: string locator
        """
        if locator:
            self.control_window(locator)
        if not self.ctx.window:
            raise WindowControlError("There is no active window")
        if not hasattr(self.ctx.window, "Minimize"):
            self.logger.warning(
                "Control '%s' does not have attribute Minimize" % self.ctx.window
            )
            return
        self.ctx.window.Minimize()

    @keyword(tags=["window"])
    def maximize_window(self, locator: str = None) -> None:
        """Maximize window defined by the locator string or the
        current active window.

        :param locator: string locator
        """
        if locator:
            self.control_window(locator)
        if not self.ctx.window:
            raise WindowControlError("There is no active window")
        if not hasattr(self.ctx.window, "Maximize"):
            raise WindowControlError("Window does not have attribute Maximize")
        self.ctx.window.Maximize()

    @keyword(tags=["window"])
    def list_windows(self) -> List[Dict]:
        """List all window Controls on the system.

        :return: list of dictionaries containing information
         about Window controls
        """
        windows = auto.GetRootControl().GetChildren()
        process_list = utils.get_process_list()
        win_list = []
        for win in windows:
            pid = win.ProcessId
            info = {
                "title": win.Name,
                "pid": win.ProcessId,
                "name": process_list[pid] if pid in process_list.keys() else None,
            }
            win_list.append(info)
        return win_list

    @keyword(tags=["window"])
    def windows_run(self, text: str, wait_time: float = 3.0) -> None:
        """Use Windows run window to launch application

        :param text: text to enter into run input field
        :param wait_time: sleep time after search has been entered (default 3.0 seconds)
        """
        self.ctx.send_keys("{Win}r")
        self.ctx.send_keys(text)
        self.ctx.send_keys("{Enter}")
        time.sleep(wait_time)

    @keyword(tags=["window"])
    def windows_search(self, text: str, wait_time: float = 3.0) -> None:
        """Use Windows search window to launch application

        :param text: text to enter into search input field
        :param wait_time: sleep time after search has been entered (default 3.0 seconds)
        """
        self.ctx.send_keys("{Win}s")
        self.ctx.send_keys(text)
        self.ctx.send_keys("{Enter}")
        time.sleep(wait_time)

    @keyword(tags=["window"])
    def close_current_window(self) -> None:
        """Closes current active window or log warning message"""
        if not self.ctx.window:
            self.ctx.logger.warning("There is no active window")
            return
        self.ctx.logger.warning(
            "Current window process id = %s" % self.ctx.window.ProcessId
        )
        self.ctx.window.SetActive()
        self.ctx.window.SendKeys("{Alt}{F4}")



from datetime import datetime

# pylint: disable=C0415
from pynput_robocorp import mouse, keyboard

from RPA.Windows.keywords import keyword, LibraryContext
from RPA.Windows import utils


if utils.is_windows():
    import uiautomation as auto


class RecorderKeywords(LibraryContext):
    """Keywords for inspecting and recording"""

    def __init__(self, ctx):
        super().__init__(ctx)
        self.record = False
        self.recording = []
        self.record_time = None

    @keyword(tags=["recording"])
    def inspect_control(self, action: str = "Click", control_window: bool = True):
        """Inspect Windows Control under mouse pointer

        :param action: which action is attached to locator
        :param control_window: set False to not include ``Control Window`` keyword
        """
        # TODO. Add Python syntax
        with auto.UIAutomationInitializerInThread(debug=True):
            control = auto.ControlFromCursor()
            self.ctx.logger.warning(control)
            parent_control = control.GetParentControl()
            top_level_control = control.GetTopLevelControl()
            # self.logger.warning("Top: %s" % top_level_control)
            # self.logger.warning("Parent: %s" % parent_control)
            # self.logger.warning(dir(control))
            parent_locator = self._get_control_key_properties(parent_control)
            child_locator = self._get_control_key_properties(control)
            locator_path = f"{parent_locator} > {child_locator}"
            if "id:" in child_locator:
                locator_path = child_locator
            output = []
            if control_window:
                output.append(f"Control Window  {top_level_control.Name}")
            if action:
                output.append(f"{action}  {locator_path}")
            else:
                output.append(locator_path)
            if self.record:
                # skip similar locators
                # unique = True
                # for item in self.recording:
                #     if (
                #         item["top"] == top_level_control.Name
                #         and item["locator"] == locator_path
                #     ):
                #         unique = False
                #         break
                # if unique:
                self.recording.append(
                    {
                        "type": "locator",
                        "top": top_level_control.Name,
                        "top_handle": top_level_control.NativeWindowHandle,
                        "x": top_level_control,
                        "locator": locator_path,
                    }
                )
            return output

    def _get_control_key_properties(self, control, regex_limit=300):
        automation_id = control.AutomationId
        snippet = ""
        if automation_id and not utils.is_integer(automation_id):
            snippet = f"id:{automation_id}"
        else:
            control_type = control.ControlTypeName.strip()
            class_name = control.ClassName.strip()
            name = control.Name.strip()
            name_property = "name:"
            if len(name) > regex_limit:
                name_property = "regex:"
                name = name[:regex_limit].strip()
            locators = []
            if len(control_type) > 0:
                locators.append(f"type:{control_type}")
            if len(class_name) > 0:
                locators.append(f"class:{class_name}")
            if len(name) > 0 and " " in name:
                locators.append(f"{name_property}'{name}'")
            elif len(name) > 0:
                locators.append(f"{name_property}{name}")
            if not locators:
                self.ctx.logger.warning(
                    "Was unable to construct locator for the control"
                )
                return None
            else:
                snippet = " and ".join(locators)
        return snippet

    @keyword(tags=["recording"])
    def stop_listeners(self):
        mouse.Listener.stop()
        keyboard.Listener.stop()

    @keyword(tags=["recording"])
    def stop_recording(self):
        self.record = False

    @keyword(tags=["recording"])
    def start_recording(self):
        """Start recording mouse clicks.

        Can be stopped by pressing keyboard ``ESC``.
        """
        self.record = True
        self.recording = []
        self.record_time = None

        def on_click(x, y, button, pressed):  # pylint: disable=W0613
            if pressed:
                inspect_time = datetime.now()
                if self.record_time:
                    timediff = inspect_time - self.record_time
                    seconds = max(
                        round(float(timediff.microseconds / 1000000.0), 1), 0.1
                    )
                    self.recording.append({"type": "sleep", "value": seconds})
                self.record_time = inspect_time
                self.inspect_control()

        def on_release(key):
            if key == keyboard.Key.esc:
                return False
            return True

        mouse_listener = mouse.Listener(on_click=on_click)
        self.ctx.logger.warning("starting mouse listener")
        mouse_listener.start()
        self.ctx.logger.warning("starting key listener")
        with keyboard.Listener(on_release=on_release) as key_listener:
            key_listener.join()
        mouse_listener.stop()
        self.ctx.logger.warning("LEN = %s" % len(self.recording))

    @keyword(tags=["recording"])
    def get_recording(self, sleeps=True):
        """Get list of recorded steps.

        :param sleeps: set False to exclude recording sleeps
        """
        # TODO. atm will always use CLICK
        output = []
        top = None
        for item in self.recording:
            if sleeps and item["type"] == "sleep":
                output.append(f"Sleep   {item['value']}s")
            if (
                item["type"] == "locator"
                and not top
                or "top" in item.keys()
                and item["top"] != top
            ):
                output.append(
                    f"Control Window    {item['top']}  # Handle: {item['top_handle']}"
                )
                top = item["top"]
            if item["type"] == "locator":
                output.append(f"Click   {item['locator']}")

        result = "\n".join(output)
        header = (
            f"\n{'-'*80}"
            "\nCOPY & PASTE BELOW CODE INTO *** Tasks *** or *** Keywords ***"
            f"\n{'-'*80}\n\n"
        )
        footer = f"\n\n{'-'*80}"
        return f"{header}{result}{footer}"
